// Persistence of Vision Ray Tracer Scene Description File
// File: meshes.inc
// Vers: 3.5
// Desc: A macro-collection for mesh-generation and enhancement
// Date: 03.28.02
// Auth: Tim Nikias Wenclawiak

//Should be overwritten, but are "defaulted":
//  _Normal_Addition_Macro(Float,Float)

//Macro for addition
#macro _Normal_Addition_Macro(X,Z) (X-Z) #end


#declare _meshes_inc_tnw=1;

/*************/
/**Make_Mesh**/
/*************/
//Given a 2D-Array (mesh-field), this returns a mesh-object
#macro Make_Mesh(_data_array)
 #local _dX=dimension_size(_data_array,1);
 #local _dZ=dimension_size(_data_array,2);
 mesh{
 //Mesh with triangles
  #local A=0;
  #while (A<_dX-1)
   #local B=0;
   #while (B<_dZ-1)
    //Shortest connection
    #if (
        vlength(_data_array[A][B]-_data_array[A+1][B+1]) >=
        vlength(_data_array[A][B+1]-_data_array[A+1][B]) )
     triangle{_data_array[ A ][B+1],_data_array[A+1][ B ],_data_array[ A ][ B ]}
     triangle{_data_array[ A ][B+1],_data_array[A+1][ B ],_data_array[A+1][B+1]}
    #else
     triangle{_data_array[A+1][B+1],_data_array[ A ][ B ],_data_array[ A ][B+1]}
     triangle{_data_array[A+1][B+1],_data_array[ A ][ B ],_data_array[A+1][ B ]}
    #end
   #local B=B+1;
   #end
  #local A=A+1;
  #end
 }
#end

/********************/
/**Make_Smooth_Mesh**/
/********************/
//Given a 2D-Array (mesh-field), this returns a smooth-mesh-object
#macro Make_Smooth_Mesh(_data_array,_normal_array)
 #local _dX=dimension_size(_data_array,1);
 #local _dZ=dimension_size(_data_array,2);
 //Mesh with triangles
 mesh{
 #local A=0;
 #while (A<_dX-1)
  #local B=0;
  #while (B<_dZ-1)
    //Shortest connection
    #if (
        vlength(_data_array[A][B]-_data_array[A+1][B+1]) >=
        vlength(_data_array[A][B+1]-_data_array[A+1][B]) )
     smooth_triangle{
       _data_array[ A ][B+1],_normal_array[ A ][B+1],
       _data_array[A+1][ B ],_normal_array[A+1][ B ],
       _data_array[ A ][ B ],_normal_array[ A ][ B ]
       }

     smooth_triangle{
       _data_array[ A ][B+1],_normal_array[ A ][B+1],
       _data_array[A+1][ B ],_normal_array[A+1][ B ],
       _data_array[A+1][B+1],_normal_array[A+1][B+1]
       }
    #else

     smooth_triangle{
       _data_array[A+1][B+1],_normal_array[A+1][B+1],
       _data_array[ A ][ B ],_normal_array[ A ][ B ],
       _data_array[ A ][B+1],_normal_array[ A ][B+1]
       }

     smooth_triangle{
       _data_array[A+1][B+1],_normal_array[A+1][B+1],
       _data_array[ A ][ B ],_normal_array[ A ][ B ],
       _data_array[A+1][ B ],_normal_array[A+1][ B ]
       }
    #end
  #local B=B+1;
  #end
 #local A=A+1;
 #end
 }
#end

/*************/
/**Make_Grid**/
/*************/
//Given a 2D-Array (mesh-field) and a thickness-float, this returns a bunch of cylinders and spheres
#macro Make_Grid(_Data_Array,_thickness)
 #local _X=dimension_size(_Data_Array,1);
 #local _Z=dimension_size(_Data_Array,2);
 
#local A=0;
#while (A<_X-1)
 #local B=0;
 #while (B<_Z-1)
 
  sphere{_Data_Array[A][B],_thickness}
  sphere{_Data_Array[A+1][B],_thickness}
  sphere{_Data_Array[A][B+1],_thickness}
  sphere{_Data_Array[A+1][B+1],_thickness}

  cylinder{_Data_Array[A][B],_Data_Array[A+1][B],_thickness}
  cylinder{_Data_Array[A+1][B],_Data_Array[A][B+1],_thickness}
  cylinder{_Data_Array[A][B+1],_Data_Array[A][B],_thickness}
  cylinder{_Data_Array[A][B+1],_Data_Array[A+1][B+1],_thickness}
  cylinder{_Data_Array[A][B+1],_Data_Array[A+1][B],_thickness}
  cylinder{_Data_Array[A+1][B+1],_Data_Array[A+1][B],_thickness}
    

 #local B=B+1;
 #end
#local A=A+1;
#end


#end

/****************/
/**Make_Normals**/
/****************/
//Given 2D-Array (mesh-field), returns 2D-Array with normals for the mesh
#macro Make_Normals(_dat_array)

 #local _X=dimension_size(_dat_array,1);
 #local _Z=dimension_size(_dat_array,2);
 #local Normal_Array=array[_X][_Z]

 //Initialize Mesh
 #local A=0;
 #while (A<_X)
  #local B=0;
  #while (B<_Z)
   #local Normal_Array[A][B]=<0,0,0>; 
  #local B=B+1;
  #end
 #local A=A+1;
 #end

 //Generate and add all normals
 #local A=0;
 #while (A<_X-1)
  #local B=0;
  #while (B<_Z-1)

   #local Mid_Sec = (_dat_array[A][B]+_dat_array[A+1][B]+_dat_array[A][B+1]+_dat_array[A+1][B+1])/4;
   
   #local Nor_1=vcross(_dat_array[A+1][B+1]-Mid_Sec,_dat_array[A][B+1]-Mid_Sec);
   #local Nor_2=vcross(_dat_array[A][B+1]-Mid_Sec,_dat_array[A][B]-Mid_Sec);
   #local Nor_3=vcross(_dat_array[A][B]-Mid_Sec,_dat_array[A+1][B]-Mid_Sec);
   #local Nor_4=vcross(_dat_array[A+1][B]-Mid_Sec,_dat_array[A+1][B+1]-Mid_Sec);
   
   #local Normal_Array[A][B]=Normal_Array[A][B]+Nor_2+Nor_3;
   #local Normal_Array[A+1][B]=Normal_Array[A][B]+Nor_3+Nor_4;
   #local Normal_Array[A][B+1]=Normal_Array[A][B]+Nor_1+Nor_2;
   #local Normal_Array[A+1][B+1]=Normal_Array[A][B]+Nor_4+Nor_1;
      
  #local B=B+1;
  #end 
 #local A=A+1;
 #end
 
 //Normalize the entire data
 #local A=0;
 #while (A<_X)
  #local B=0;
  #while (B<_Z)
   #local Normal_Array[A][B]=vnormalize(Normal_Array[A][B]); 
  #local B=B+1;
  #end
 #local A=A+1;
 #end

Normal_Array

#end

/***************/
/**Add_Normals**/
/***************/
//Given 2D-Array (Mesh-Field), normals are calculated and added with macro
#macro Add_Normals(_dat_array,_amount)
 #local _X=dimension_size(_dat_array,1);
 #local _Z=dimension_size(_dat_array,2);
 
 #local _n_data=Make_Normals(_dat_array)
 
 #local A=0;
 #while (A<_X)
  #local B=0;
  #while (B<_Z)
   
   #declare _dat_array[A][B]=_dat_array[A][B]+_n_data[A][B]*_Normal_Addition_Macro( A/(_X-1), B/(_Z-1) )*_amount; 

  #local B=B+1;
  #end
 #local A=A+1;
 #end
 
#end

/*******************/
/*Mirror_Mesh_Dat*/
/*******************/
//Given 2D-Array (Mesh-field), and mirror-vector the field is enlarged with mirroring at X-Axis
#macro Mirror_Mesh_Dat(Node_Field,_MVec)
 #local _X=dimension_size(Node_Field,1);
 #local _Z=dimension_size(Node_Field,2);
 
 #local Offset=on;
 //Check if we are directly on axis, if not...
 #if (_MVec.x=1)
  #if (Node_Field[0][0].x!=0)
   #local Offset=off;
   #end 
 #end

 #if (_MVec.y=1)
  #if (Node_Field[0][0].y!=0)
   #local Offset=off;
  #end 
 #end

 #if (_MVec.z=1)
  #if (Node_Field[0][0].z!=0)
   #local Offset=off;
  #end 
 #end
 
 #if (Offset)
 #local _NF=array[_X*2-1][_Z]
 #local A=0;
 #while (A<_X)
  #local B=0;
  #while (B<_Z)
    
    #local _NF[A][B]=Node_Field[_X-A-1][B]*(<1,1,1>-2*_MVec);
    #local _NF[A+_X-1][B]=Node_Field[A][B];

  #local B=B+1;
  #end
 #local A=A+1;
 #end
 #else
 #local _NF=array[_X*2][_Z]
 #local A=0;
 #while (A<_X)
  #local B=0;
  #while (B<_Z)
    
    #local _NF[A][B]=Node_Field[_X-A-1][B]*(<1,1,1>-2*_MVec);
    #local _NF[A+_X][B]=Node_Field[A][B];

  #local B=B+1;
  #end
 #local A=A+1;
 #end
 #end 
  
#declare Node_Field=_NF

#end


/********************/
/*Translate_Mesh_Dat*/
/********************/
//Given 2D-Array (Mesh-field) and translation, given array is altered
#macro Translate_Mesh_Dat(Node_Field,_MVec)
 #local _X=dimension_size(Node_Field,1);
 #local _Z=dimension_size(Node_Field,2);
 //Move through entire Data
 #local A=0;
 #while (A<_X)
  #local B=0;
  #while (B<_Z)
    #declare Node_Field[A][B]=Node_Field[A][B]+_MVec;
  #local B=B+1;
  #end
 #local A=A+1;
 #end
#end

/********************/
/*Rotate_Mesh_Dat*/
/********************/
//Given 2D-Array (Mesh-field) and translation, given array is altered
#macro Rotate_Mesh_Dat(Node_Field,_RVec)
 #local _X=dimension_size(Node_Field,1);
 #local _Z=dimension_size(Node_Field,2);
 //Move through entire Data
 #local A=0;
 #while (A<_X)
  #local B=0;
  #while (B<_Z)
    #declare Node_Field[A][B]=Node_Field[A][B]+_RVec;
  #local B=B+1;
  #end
 #local A=A+1;
 #end
#end

/********************/
/*Scale_Mesh_Dat*/
/********************/
//Given 2D-Array (Mesh-field) and translation, given array is altered
#macro Scale_Mesh_Dat(Node_Field,_SVec)
 #local _X=dimension_size(Node_Field,1);
 #local _Z=dimension_size(Node_Field,2);
 //Move through entire Data
 #local A=0;
 #while (A<_X)
  #local B=0;
  #while (B<_Z)
    #declare Node_Field[A][B]=Node_Field[A][B]*_SVec;
  #local B=B+1;
  #end
 #local A=A+1;
 #end
#end
